
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : htmltm-test.scm
;; DESCRIPTION : Test suite for htmltm
;; COPYRIGHT   : (C) 2002  David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert htmltm htmltm-test)
  (:use (convert htmltm htmltm) (convert tools xmltm))
  (:export regtest-htmltm))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Markup
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (regtest-htmltm-grouping)
  (regression-test-group
   "htmltm, grouping markup" "grouping"
   htmltm-as-serial :none
   (test "div" '(div "a") '(document "a"))
   (test "span" '(span "a") "a")))

(define (regtest-htmltm-headings)
  (regression-test-group
   "htmltm, heading markup" "headings"
   htmltm-as-serial :none
   (test "h1" '(h1 "a") '(document (chapter "a")))
   (test "h2" '(h2 "a") '(document (section "a")))
   (test "h3" '(h3 "a") '(document (subsection "a")))
   (test "h4" '(h4 "a") '(document (subsubsection "a")))
   (test "h5" '(h5 "a") '(document (paragraph "a")))
   (test "h6" '(h6 "a") '(document (subparagraph "a")))))

(define (regtest-htmltm-address-bdo)
  (regression-test-group
   "htmltm, address and text direction" "address-bdo"
   htmltm-as-serial :none
   (test "address" '(address "a") '(document "a"))
   (test "bdo" '(bdo "a") "a")))

(define (regtest-htmltm-phrase)
  (regression-test-group
   "htmltm, phrase markup" "phrase"
   htmltm-as-serial :none
   (test "em" '(em "a") '(em "a"))
   (test "strong" '(strong "a") '(strong "a"))
   (test "cite" '(cite "a") '(cite* "a"))
   (test "dfn" '(dfn "a") '(dfn "a"))
   (test "code" '(code "a") '(code* "a"))
   (test "samp" '(samp "a") '(samp "a"))
   (test "kbd" '(kbd "a") '(kbd "a"))
   (test "var" '(var "a") '(var "a"))
   (test "abbr" '(abbr "a") '(abbr "a"))
   (test "acronym" '(acronym "a") '(acronym "a"))))

(define (regtest-htmltm-quotation)
  (regression-test-group
   "htmltm, quotation markup" "quotation"
   htmltm-as-serial :none
   (test "blockquote, empty"
	 '(blockquote) '(document (quotation (document ""))))
   (test "blockquote, one para"
	 '(blockquote "a") '(document (quotation (document "a"))))
   (test "blockquote, two paras"
	 '(blockquote "a" (p "b")) '(document (quotation (document "a" "b"))))
   (test "inline quote" '(q "a")  "``a''")))

(define (regtest-htmltm-subsuper)
  (regression-test-group
   "htmltm, subscripts and superscripts" "subsuper"
   htmltm-as-serial :none
   (test "superscript" '(sup "a") '(rsup "a"))
   (test "subscript" '(sub "a") '(rsub "a"))))

(define (regtest-htmltm-para)
  (+ (regtest-htmltm-para-normal) (regtest-htmltm-preformatted)))

(define (regtest-htmltm-para-normal)
  (regression-test-group
   "htmltm, lines and paragraphs" "para"
   htmltm-as-serial :none
   (test "p, empty" '(p) '(document ""))
   (test "two p, empty" '(html (p) (p)) '(document "" ""))
   (test "p, text" '(p "a") '(document "a"))
   (test "p text, p empty, p text"
	 '(html (p "a") (p) (p "b")) '(document "a" "" "b"))
   (test "br" '(br) '(format "next line"))
   (test "br in p"
	 '(p "a" (br) "b") '(document (concat "a" (format "next line") "b")))))

(define (regtest-htmltm-preformatted)
  (define (code l) `(document (code (document ,@l))))
  (regression-test-group
   "htmltm, preformatted text" "pre"
   htmltm-as-serial code
   (test "empty" '(pre) '(""))
   (test "emtpy, newlines" '(pre "\n\n") '(""))
   (test "one line" '(pre "hello") '("hello"))
   (test "one line, whitespace" '(pre " hello ") '(" hello "))
   (test "one line, newlines" '(pre "\n hello \n") '(" hello "))
   (test "two lines" '(pre "hello\nworld") '("hello" "world"))
   (test "two lines, whitespace"
	 '(pre " hello \n world ") '(" hello " " world "))
   (test "two lines, newlines"
	 '(pre "\n hello \n world \n") '(" hello " " world "))
   (test "inline tags"
	 '(pre (em " hello ")) '((em " hello ")))
   (test "inline tags, newlines"
	 '(pre "\n" (em " hello ") "\n") '((em " hello ")))))

(define (regtest-htmltm-changes)
  (regression-test-group
   "htmltm, document changes" "changes"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-list-items)
  (define (make-itemize l)
    `(document (itemize (document ,@l))))
  (regression-test-group
   "htmltm, lists items" "list-items"
   htmltm-as-serial make-itemize
   (test "empty list" '(ul) '(""))
   (test "one empty item" '(ul (li)) '((item)))
   (test "two empty items" '(ul (li) (li)) '((item) (item)))
   (test "one item inline"
	 '(ul (li "a")) '((concat (item) "a")))
   (test "one item, id, inline"
	 '(ul (li (@ (id "x")) "a")) '((concat (item) (label "x") "a")))
   (test "one item one p"
	 '(ul (li (p "a"))) '((concat (item) "a")))
   (test "one item two p"
	 '(ul (li (p "a") (p "b"))) '((concat (item) "a") "b"))
   (test "one item, inline and p"
	 '(ul (li "a" (p "b"))) '((concat (item) "a") "b"))
   (test "one item, inline and p with structure"
	 '(ul (li "a" (em "b") (p "c")))
	 '((concat (item) "a" (em "b")) "c"))
   (test "two items, inline"
	 '(ul (li "a") (li "b")) '((concat (item) "a") (concat (item) "b")))
   (test "two items, blocks"
	 '(ul (li (p "a")) (li (p "b")))
	 '((concat (item) "a") (concat (item) "b")))))

(define (regtest-htmltm-list-definition)
  (define (make-description l)
    `(document (description (document ,@l))))
  (regression-test-group
   "htmltm, definition lists" "list-definition"
   htmltm-as-serial make-description
   (test "empty list" '(dl) '(""))
   (test "empty dt" '(dl (dt)) '((item* "")))
   (test "emtpy dd" '(dl (dd)) '(""))
   (test "empty dt, dd" '(dl (dt) (dd)) '((item* "")))
   (test "empty dd, dt" '(dl (dd) (dt)) '("" (item* "")))
   (test "empty dt, dt, dd, dd"
	 '(dl (dt) (dt) (dd) (dd)) '((item* "") (item* "") ""))
   (test "dt, dt, dd, dd with text"
	 '(dl (dt "a") (dt "b") (dd "c") (dd "d"))
	 '((item* "a") (concat (item* "b") "c") "d"))))
   
(define (regtest-htmltm-list-kinds)
  (regression-test-group
   "htmltm, other kinds of lists" "list-kinds"
   htmltm-as-serial :none
   (test "enumeration" '(ol (li "a"))
	 '(document (enumerate (document (concat (item) "a")))))))

(define (regtest-htmltm-lists)
  (+ (regtest-htmltm-list-items)
     (regtest-htmltm-list-definition)
     (regtest-htmltm-list-kinds)))

(define (regtest-htmltm-table)
  (regression-test-group
   "htmltm, table" "table"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-links)
  (regression-test-group
   "htmltm, links" "links"
   htmltm-as-serial :none
   (test "name" '(a (@ (name "foo")) "bar") '(concat (label "foo") "bar"))
   (test "id" '(a (@ (id "foo")) "bar") '(concat (label "foo") "bar"))
   (test "href" '(a (@ (href "foo")) "bar") '(hlink "bar" "foo"))
   (test "name and href" '(a (@ (name "foo") (href "bar")) "baz")
	 '(hlink (concat (label "foo") "baz") "bar"))
   (test "id and href" '(a (@ (id "foo") (href "bar")) "baz")
	 '(hlink (concat (label "foo") "baz") "bar"))
   (test "id, name and href"
	 '(a (@ (id "eggs") (name "foo") (href "bar")) "baz")
	 '(hlink (concat (label "eggs") (label "foo") "baz") "bar"))))

(define (regtest-htmltm-objects)
  (define (postscript s w h) `(postscript ,s ,w ,h "" "" "" ""))
  (regression-test-group
   "htmltm, object images and applets" "objects"
   htmltm-as-serial :none
   (test "img" '(img (@ (src "foo"))) (postscript "foo" "*6383/10000" ""))
   (test "img, px px"
	 '(img (@ (src "foo") (width "100") (height "50")))
	 (postscript "foo" "100px" "50px"))
   (test "img, %"
	 '(img (@ (src "foo") (width "50%")))
	 (postscript "foo" "0.5par" ""))))

(define (regtest-htmltm-alignement)
  (regression-test-group
   "htmltm, alignement" "alignement"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-style)
  (regression-test-group
   "htmltm, font style" "style"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-font)
  (regression-test-group
   "htmltm, font modifiers" "font"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-rules)
  (regression-test-group
   "htmltm, horizontal rules" "rules"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-frames)
  (regression-test-group
   "htmltm, frames" "frames"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-forms)
  (regression-test-group
   "htmltm, forms" "forms"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-scripting)
  (regression-test-group
   "htmltm, scripting" "scripting"
   htmltm-as-serial :none
   ))

(define (regtest-htmltm-handlers-extra)
  ;; Maybe those tests should be split in actual unit tests. This extra
  ;; handling is so complicated it might well work for a wrong reason.
  (+ (regtest-htmltm-text-cleaning)
     (regtest-htmltm-trimming)
     (regtest-htmltm-ids)))

(define (regtest-htmltm-text-cleaning)
  (regression-test-group
   "htmltm, extra features of handlers" "handlers"
   htmltm-as-serial :none
   ;; TODO: test limitations in cleaning in preformatted text.
   (test "mixed space"
	 '(em " a\n  " (strong " b ") "\n ") '(em (concat "a " (strong "b"))))
   (test "text in element content"
	 '(ul "foo" (li "hello") "bar" (li "world") "baz")
	 '(document (itemize (document (concat (item) "hello")
				       (concat (item) "world")))))))

(define (regtest-htmltm-trimming)
  (define nl '(format "next line"))
  (define (id s) `(a (@ (id ,s))))
  (define (make-itemize l) `(document (itemize (document ,@l))))
  (regression-test-group
   ;; TODO: test disabling of trimming in preformatted text.
   "htmltm, trimming converted nodes" "trimming"
   htmltm-as-serial :none
   (test "block after"
	 '(p "a " (p "b")) '(document "a" "b"))
   (test "block before"
	 '(p (p "a") " b") '(document "a" "b"))
   (test "blocks across"
	 '(p "a " (p "b") " c " (p "d") " e")
	 '(document "a" "b" "c" "d" "e"))
   (test "white inlines between blocks"
	 '(p " " (p "a") " " (p "b") " ")
	 '(document "a" "b"))

   (test "block after, label"
	 `(p "a " ,(id "x") (p "b"))
	 '(document (concat "a" (label "x")) "b"))
   (test "block before, label"
	 `(p (p "a") ,(id "x") " b")
	 '(document "a" (concat (label "x") "b")))
   (test "blocks across, labels"
	 `(p "a " ,(id "x") (p "b")
	     ,(id "y") " c " ,(id "z") (p "d") ,(id "t") " e")
	 '(document (concat "a" (label "x")) "b"
		    (concat (label "y") "c" (label "z")) "d"
		    (concat (label "t") "e")))
   (test "block after, padded label"
	 `(p "a " ,(id "x") " " (p "b"))
	 '(document (concat "a" (label "x")) "b"))
   (test "block before, padded label"
	 `(p (p "a") " " ,(id "x") " b")
	 '(document "a" (concat (label "x") "b")))
   (test "blocks across, padded labels"
	 `(p "a " ,(id "x") " " (p "b")
	     " " ,(id "y") " c " ,(id "z") " " (p "d") " " ,(id "t") " e")
	 '(document (concat "a" (label "x")) "b"
		    (concat (label "y") "c" (label "z")) "d"
		    (concat (label "t") "e")))

   (test "newline after"
	 '(p "a " (br)) `(document (concat "a" ,nl)))
   (test "newline before"
	 '(p (br) " b") `(document (concat ,nl "b")))
   (test "newlines across"
	 '(p "a " (br) " b " (br) " c")
	 `(document (concat "a" ,nl "b" ,nl "c")))
   (test "newline in list"
	 '(ul (li " hello " (br) " world "))
	 (make-itemize `((concat (item) "hello" ,nl "world"))))

   (test "newline after, label"
	 `(p "a " ,(id "x") (br))
	 `(document (concat "a" (label "x") ,nl)))
   (test "newline before, label"
	 `(p (br) ,(id "x") " b") `(document (concat ,nl (label "x") "b")))
   (test "newlines across, labels"
	 `(p "a ",(id "x")  (br) ,(id "y") " b " ,(id "z") (br) ,(id "t") " c")
	 `(document (concat "a" (label "x") ,nl
			    (label "y") "b" (label "z") ,nl (label "t") "c")))
   (test "newline after, padded label"
	 `(p "a " ,(id "x") " " (br))
	 `(document (concat "a" (label "x") ,nl)))
   (test "newline before, padded label"
	 `(p (br) " " ,(id "x") " b") `(document (concat ,nl (label "x") "b")))
   (test "newlines across, padded labels"
	 `(p "a ",(id "x") " " (br)
	     " " ,(id "y") " b " ,(id "z") " " (br) " " ,(id "t") " c")
	 `(document (concat "a" (label "x") ,nl (label "y") "b" (label "z") ,nl
			    (label "t") "c")))))
   
(define (regtest-htmltm-ids)
  (regression-test-group
   "htmltm, conversion of id attributes" "ids"
   htmltm-as-serial :none
   ;; TODO: Test deep insertion of labels, and correct behaviour with list
   ;; markers and section titles.
   (test "string" '(em (@ (id "x")) "y") '(em (concat (label "x") "y")))
   (test "concat"
	 '(em (@ (id "x")) "y" (strong "z"))
	 '(em (concat (label "x") "y" (strong "z"))))
   ;; TODO: test insertion in para when supported"
   (test "document"
	 '(p (@ (id "x")) "y") '(document (concat (label "x") "y")))
   ;; TODO: test insertion in explicit expand documents
   (test "with document"
	 '(center (@ (id "x")) "y")
	 '(document (with "paragraph mode" "center"
			  (document (concat (label "x") "y")))))
   (test "itemize 1"
	 '(ul (@ (id "x")) (li "y"))
	 '(document (itemize (document (concat (item) (label "x") "y")))))
   (test "itemize 2"
	 '(ul (@ (id "x")) (li (@ (id "y")) "z"))
	 '(document (itemize (document
			      (concat (item) (label "x") (label "y") "z")))))
   (test "description"
	 '(dl (@ (id "x")) (dt "y") (dd "z"))
	 '(document (description
		     (document
		      (concat (item* (concat (label "x") "y")) "z")))))
   (test "section"
	 '(center (@ (id "x")) (h1 "y"))
	 '(document (with "paragraph mode" "center"
			  (document (concat (chapter "y") (label "x"))))))
   (test "section 2"
	 '(center (@ (id "x")) (h1 (@ (id "y")) "z"))
	 '(document (with "paragraph mode" "center"
			  (document (concat (chapter "z")
					    (label "x") (label "y"))))))
   (test "implicit expand document"
	 '(pre (@ (id "x")) "y")
	 '(document (code (document (concat (label "x") "y")))))
   (test "otherwise"
	 '(em (@ (id "x")) (strong "y"))
	 '(em (concat (label "x") (strong "y"))))))

(define (regtest-htmltm-less-gtr)
  (define (html->tm s) (htmltm-as-serial (parse-html s)))
  (define html1 "&lt;x&gt;")
  (define stm1 "<less>x<gtr>")
  (define stm-label '(label "<less>y<gtr>"))
  ;; Be sure that angle brackets are escaped only once 
  (regression-test-group
   "htmltm, escaping less and gtr" "less-gtr"
   html->tm :none
   (test "string node" html1 stm1)
   (test "tag with id"
	 (string-append "<em id='<y>'>" html1 "</em>")
	 `(em (concat ,stm-label ,stm1)))
   (test "anchor with id"
	 (string-append "<a id='<y>'>" html1 "</a>")
	 `(concat ,stm-label ,stm1))
   (test "anchor with name"
	 (string-append "<a name='<y>'>" html1 "</a>")
	 `(concat ,stm-label ,stm1))
   (test "anchor with href"
	 (string-append "<a href='<z>'>" html1 "</a>")
	 `(hlink ,stm1 "<less>z<gtr>"))
   (test "image source"
	 "<img src='<z>'>"
	 '(postscript "<less>z<gtr>" "*6383/10000" "" "" "" "" ""))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Test suite
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (regtest-htmltm)
  (let ((n (+ (regtest-htmltm-grouping)
	      (regtest-htmltm-headings)
	      (regtest-htmltm-address-bdo)
	      (regtest-htmltm-phrase)
	      (regtest-htmltm-quotation)
	      (regtest-htmltm-subsuper)
	      (regtest-htmltm-para)
	      (regtest-htmltm-changes)
	      (regtest-htmltm-lists)
	      (regtest-htmltm-table)
	      (regtest-htmltm-links)
	      (regtest-htmltm-objects)
	      (regtest-htmltm-alignement)
	      (regtest-htmltm-style)
	      (regtest-htmltm-font)
	      (regtest-htmltm-rules)
	      (regtest-htmltm-frames)
	      (regtest-htmltm-forms)
	      (regtest-htmltm-scripting)
	      (regtest-htmltm-handlers-extra)
	      (regtest-htmltm-less-gtr))))
    (display* "Total: " (object->string n) " tests.\n")
    (display "Test suite of htmltm: ok\n")))
