
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : htmltm.scm
;; DESCRIPTION : conversion of Html trees to TeXmacs trees
;; COPYRIGHT   : (C) 2003  David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert htmltm htmltm)
  (:use (convert tools tmlength) (convert tools sxml) (convert tools stm)
	(convert tools xmltm) (convert mathtm mathtm))
  (:export html->tree html->document
	   htmltm-as-serial ;; for htmltm-test
	   ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Specific markup
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (htmltm-list-item env a c)
  ;; List markers are glued by the list handler.
  (list (xmltm-label-decorate
	 a 'id (htmltm-serial (htmltm-preserve-space? env)
			      `((document (item)) ,@(htmltm-args env c))))))

(define (htmltm-quote env a c)
  ;; WARNING: this is incomplete. Texmacs should have a macro for inline
  ;; quotation which puts in quotation marks in a language sensitive manner.
  (list (xmltm-label-decorate
	 a 'id (htmltm-serial (htmltm-preserve-space? env)
			      `("``" ,@(htmltm-args env c) "''")))))

(define (htmltm-anchor env a c)
  (list (htmltm-href->hlink
	 a (xmltm-label-decorate
	    a 'id (xmltm-label-decorate
		   a 'name (htmltm-args-serial env c))))))

(define (htmltm-href->hlink a body)
  (let ((href (shtml-attr-non-null a 'href)))
    (if href `(hlink ,body ,(htmltm-string href)) body)))

(define (htmltm-dimension attrs name)
  (let ((s (shtml-attr-non-null attrs name)))
    (if (not s) (tmlength)
	(cond ((string-null? s) (tmlength))
	      ((string->number s) => (lambda (n) (tmlength n 'px)))
	      ((and (string-ends? s "%")
		    (/ (string->number (string-drop-right s 1)) 100))
	       => (lambda (n) (tmlength n 'par)))
	      (else (tmlength))))))

(define (htmltm-image env a c)
  (let* ((s (htmltm-string (or (shtml-attr-non-null a 'src) "")))
	 (w (tmlength->string (htmltm-dimension a 'width)))
	 (h (tmlength->string (htmltm-dimension a 'height))))
    (list (xmltm-label-decorate
	   a 'id
	   (if (not (and (string-null? w) (string-null? h)))
	       `(postscript ,s ,w ,h "" "" "" "")
	       `(postscript ,s "*6383/10000" "" "" "" "" ""))))))

(define (htmltm-font env a c)
  ;; WARNING: do as old filter, but is fragile and not conformant
  (list (htmltm-with-size
	 a (htmltm-with-color
	    a (xmltm-label-decorate
	       a 'id (htmltm-args-serial env c))))))

(define (htmltm-with-size a x)
  (let* ((sz (shtml-attr-non-null a 'size))
	 (mult
	  (assoc sz '(("-4" "0.5") ("-3" "0.6") ("-2" "0.7") ("-1" "0.8")
		      ("+1" "1.2") ("+2" "1.4") ("+3" "1.7") ("+4" "2")))))
    (if mult `(with "font size" ,(second mult) ,x) x)))

(define (htmltm-with-color a x)
  (let* ((col (shtml-attr-non-null a 'color)))
    (if col `(with "color" ,(htmltm-string col) ,x) x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main translation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (htmltm-drop env a c) '())

(define (htmltm-pass env a c)
  (let ((l (htmltm-args env c)))
    (if (and (null? l) (not (assoc 'id a))) '()
	(list (xmltm-label-decorate
	       a 'id (htmltm-serial
		      (htmltm-preserve-space? env) l))))))

(define (htmltm-args env l)
  ;; Convert the content list of an HTML element.
  (append-map (lambda (x) (htmltm env x)) l))

(define (htmltm-args-serial env l)
  (htmltm-serial (htmltm-preserve-space? env) (htmltm-args env l)))

(define (htmltm env t)
  ;; Convert a shxml element to texmacs.
  ;; Return a list which is either null, or contains a single serial node.
  ;; All methods must use this return convention.
  (sxml-dispatch
   (lambda (env t) (list (htmltm-string t)))
   htmltm-pass "h" env t))

(define (htmltm-as-serial t)
  ;; As htmltm, but returns a serial node
  (with-htmltm-environment env
    (htmltm-serial (htmltm-preserve-space? env) (htmltm env t))))

(define handler (cut htmltm-handler <> <> <> htmltm-args-serial))

(drd-dispatcher htmltm-methods%
  ;;; Document structure
  ((:or head title meta) htmltm-drop)
  ((:or html body) htmltm-pass)

  ;; Grouping
  (div  (handler :mixed :block  htmltm-pass))
  ;; TODO: convert 'align' attributes in div, p and headings
  (span (handler :mixed :inline htmltm-pass))

  ;; Headings
  (h1 (handler :mixed :block "chapter"))
  (h2 (handler :mixed :block "section"))
  (h3 (handler :mixed :block "subsection"))
  (h4 (handler :mixed :block "subsubsection"))
  (h5 (handler :mixed :block "paragraph"))
  (h6 (handler :mixed :block "subparagraph"))

  ;; Address and text direction
  (address (handler :mixed :block htmltm-pass))
  (bdo (handler :mixed :inline htmltm-pass))
  
  ;;; Structured text
  ;; Phrase elements
  (em      (handler :mixed :inline "em"))
  (strong  (handler :mixed :inline "strong"))
  (cite    (handler :mixed :inline "cite*"))
  (dfn     (handler :mixed :inline "dfn"))
  (code    (handler :mixed :inline "code*"))
  (samp    (handler :mixed :inline "samp"))
  (kbd     (handler :mixed :inline "kbd"))
  (var     (handler :mixed :inline "var"))
  (abbr    (handler :mixed :inline "abbr"))
  (acronym (handler :mixed :inline "acronym"))

  ;; Quotations
  ;; NOTE: there should be a texmacs macro for Q
  (q (handler :mixed :inline htmltm-quote))
  (blockquote (handler :mixed :block "quotation"))

  ;; Subscripts and superscripts
  (sub (handler :mixed :inline '(rsub)))
  (sup (handler :mixed :inline '(rsup)))

  ;; Lines and paragraphs
  (p (handler :mixed :block htmltm-pass))
  (br (handler :empty :inline '((format "next line"))))
  (pre (handler :pre :block "code"))

  ;; Document changes
  ((:or ins del) (handler :mixed :inline htmltm-pass))

  ;;; Lists
  (dl (handler :element :block "description"))
  (dt (handler :mixed :block "item*"))
  (dd (handler :mixed :block htmltm-pass))
  (ol (handler :element :block "enumerate"))
  (ul (handler :element :block "itemize"))
  (li (handler :mixed :block htmltm-list-item))
  (menu (handler :element :block "itemize")) ; deprecated
  (dir  (handler :element :block "itemize")) ; deprecated

  ;;; Tables
  (table (handler :element :block htmltm-pass))
  ;; Elements allowed only in table
  ((:or col colgroup tbody) htmltm-pass)
  (tr (handler :element :block htmltm-pass))
  (td (lambda (env a c)
	(list (htmltm-serial
	       (htmltm-preserve-space? env)
	       `(" " ,@(htmltm-args env (htmltm-space-mixed env c)) " ")))))

  ;;; Links
  (a (handler :mixed :inline htmltm-anchor))
  ;; Elements allowed only in HEAD
  ((:or (link base)) htmltm-drop)

  ;;; Objects images and applets
  (object (handler :mixed :inline htmltm-drop))
  ;; TODO: handle cases where OBJECT is equivalent to IMG
  (param htmltm-drop) ; allowed only in object
  (img (handler :empty :inline htmltm-image))
  (applet (handler :mixed :inline htmltm-drop))
  (map (handler :element :inline htmltm-drop))
  (area htmltm-drop) ; allowed only in map

  ;;; Alignement, font styles and horizontal rules
  ;; Alignement (deprecated)
  ;; NOTE: the center macro is now deprecated. 
  (center (handler :mixed :block '(with "paragraph mode" "center")))

  ;; Font style
  (tt  (handler :mixed :inline '(with "font family" "tt")))
  ; NOTE: the tt macro is now deprecated
  (i   (handler :mixed :inline '(with "font shape" "italic")))
  (b   (handler :mixed :inline '(with "font series" "bold")))
  (big (handler :mixed :inline '(with "font size" "1.2")))
  (small (handler :mixed :inline '(with "font size" "0.83")))
  ((:or s strike) (handler :mixed :inline htmltm-pass))
  (u (handler :mixed :inline "underline"))

  ;; Font modifiers (deprecated)
  (font (handler :mixed :inline htmltm-font))
  (basefont htmltm-drop)

  ;; Rules
  (hr (handler :empty :block '((value "hrule"))))

  ;;; Frames
  (frameset htmltm-drop)
  (frame htmltm-drop) ; allowed only in FRAMESET
  (noframes htmltm-pass)
  (iframe (handler :mixed :block htmltm-pass))
  
  ;;; Forms
  (form (handler :mixed :block htmltm-drop))
  ;; elements allowed only in FORM
  ((:or input button select optgroup option textarea isindex label fieldset
	legend) htmltm-drop)

  ;;; Scripting
  (script htmltm-drop)
  (noscript (handler :mixed :block htmltm-pass))
  
  ;; Tags present in the previous converter
  ;; Unknown: FIG FN NOTE AU LANG PERSON
  ;; ABBREV instead of ABBR
  ;; BQ as a shorthand for BLOCKQUOTE
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; WARNING: I am not sure the return types are right

(tm-define (html->tree s)
  (:type (forall T (string -> (list T))))
  (:synopsis "Convert the HTML @s into a document fragment.")
  (catch #t
	 (lambda () (htmltm-as-serial (parse-html s)))
	 (lambda args (object->string args))))

(tm-define (html->document s)
  (:type (forall T (string -> (tuple (list T) string))))
  (:synopsis "Convert the HTML @s into a full document.")
  (:returns "A list of two elements:"
	    "  -- the document object;"
	    "  -- the document style as a string.")
  ;; FIXME: should also be able to specify initial environment values, like the
  ;;   document language.
  (list (stm-unary-document (html->tree s)) 'browser))
