
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : xmltm.scm
;; DESCRIPTION : Common tools to import XML data.
;; COPYRIGHT   : (C) 2002-2003  David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert tools xmltm)
  (:use (convert tools stm) (convert tools sxml))
  (:export environment environment-ref* environment-ref
	   with-environment* with-environment
	   xmltm-parse htmltm-parse
	   with-htmltm-environment htmltm-preserve-space?
	   htmltm-string shtml-attr-non-null xmltm-label-decorate
	   htmltm-space-mixed 
	   htmltm-handler mathtm-handler sxml-dispatch
	   htmltm-serial mathtm-serial))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Converter environments
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Environments are used by converters to accumulate dynamic state.
;;
;; By policy, an environment can only be mutated inside a bounded dynamic scope
;; (i.e. environment-set! must stay private).
;;
;; The intended semantics of environment is "functional operations on
;; hash-tables". However, for simplicity and efficiency a dynamic scoping
;; implementation is used.
;;
;; WARNING: environments are NOT continuation-correct. A continuation which
;; alters an environment is only garanteed to evaluate correctly _once_.
;;
;; WARNING: with-environment* does not restore the environment when it is
;; exited by stack unwinding (when an exception occurs).
;;
;; NOTE: a pure functional solution could be implemented using alists. If the
;; corner-case semantics of the hash-table based solution become problematic,
;; that should be the next thing to try. Since the list is not expected to grow
;; longer than a few dozen items, list lookup should not be a big problem.
;; Though a useful optimization may be adaptative ordering based on access
;; count.

;; Environment primitives

(define (environment)
  (make-ahash-table))

(define-macro (environment-set!* env key val) ; must stay private
  `(ahash-set! ,env ,key ,val))

(define-macro (environment-remove! env key) ; must stay private
  `(ahash-remove! ,env ,key))

(define (environment-ref* env key)
  (let ((h (ahash-get-handle env key)))
    (if h (cdr h)
	;; WARNING: guile-1.3.4 requires %S instead of ~S
	(texmacs-error "Unbound key in environment:" key))))

(define-macro (environment-ref env key)
  `(environment-ref* ,env (quote ,key)))

(define (environment-binding env key)
  ;; If @key is bound in @env returns (list @key value). Otherwise, return #f.
  ;; Unlike hash-get-handle, the return value is not mutable.
  (let ((h (ahash-get-handle env key)))
    (list key (and (pair? h) (list (cdr h))))))

;; Environment library (only use primitives)

(define (environment-bind! env binding)	; must stay private
  ;; Restore a binding previously saved with environment-binding.
  (if (second binding)      
      (environment-set!* env (first binding) (car (second binding)))
      (environment-remove! env (first binding))))

(define (with-environment* env bindings proc)
  (let ((saves '()) (result #f))
    (for-each (lambda (b)
		(set-cons! saves (environment-binding env (first b)))
		(environment-set!* env (first b) (second b)))
	      bindings)
    (set! result (proc env))
    (for-each (lambda (b)
		(environment-bind! env b))
	      saves)
    result))

(define (list-of-length-2? x)
  (and (pair? x)
       (not (null? (cdr x)))
       (null? (cddr x))))

(define-macro (with-environment env bindings . body)
  (if (not (list? bindings))
      ;; WARNING: guile-1.3.4 requires %s instead of ~A
      (syntax-error "Bindings are not a list: ~A" bindings))
  `(with-environment* ,env
       (list ,@(map-in-order
		(lambda (b)
		  (if (not (list-of-length-2? b))
		      ;; WARNING: guile-1.3.4 requires %s instead of ~A
		      (syntax-error "Ill-formed binding: ~A" b))
		  `(list (quote ,(first b)) ,(second b)))
		bindings))
    (lambda (,env) ,@body)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; XML namespace normalization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; The built-in parser xml/html parser is not namespace aware. So we need to
;; perform namespace-prefix normalization ourselves after the parser. Only
;; namespace-normalized xml trees can be reliably queried.
;;
;; Namespace-normalization deletes all "xmlns" and "xmlns:*" attributes and
;; change the name of all elements and attributes to a fixed set of normalized
;; namespace prefixes.
;;
;; Normalized namespace prefixes are:
;;   x  --  XML - http://www.w3.org/XML/1998/namespace
;;          Reserved prefix (e.g. xml:space).
;;   h  --  XHTML - http://www.w3.org/1999/xhtml
;;          Any HTML or HTML-like data.
;;   m  --  MathML - http://www.w3.org/1998/Math/MathML
;;
;; Since the parser is designed to be used for conversion to STM data format,
;; no provisions are made to preserve the namespace prefixes used in the
;; orginial sxml tree. Namespace normalization is not reversible.

(define xmlns-uri-xml "http://www.w3.org/XML/1998/namespace")
(define xmlns-uri-xhtml "http://www.w3.org/1999/xhtml")
(define xmlns-uri-mathml "http://www.w3.org/1998/Math/MathML")

;;; Building the namespace bindings environment

(define-macro (with-xmltm-environment env ns . body)
  `(let ((,env (environment)))
     (with-environment ,env ((*default* ,ns)
			     (xml ,xmlns-uri-xml))
       ,@body)))

(define (consume-xmlns env attrs proc)
  (receive (attrs f+bindings )
      ((cut list-partition <> first)
       ((cut map <> attrs)
	(lambda (attr)
	  (if (eq? 'xmlns (first attr))
	      `(#f *default* ,(second attr))
	      (receive (ns-id ncname) (sxml-split-name (first attr))
		(if (== "xmlns" ns-id)
		    (list #f (string->symbol ncname) (second attr))
		    attr))))))
    ;; Either bindings or attributes are prefixed by #f to allow easy
    ;; partition. Regular attributes are (likely) more numerous than xmlns
    ;; attributes, so bindings are prefixed to reduce prefix-stripping cost.
    (if (null? f+bindings) (proc env attrs)
	(with-environment* env (map cdr f+bindings)
	  (cute proc <> attrs)))))

;;; Converting nodes

(define (htmltm-parse s)
  (xmltm-parse xmlns-uri-xhtml s))

(define (xmltm-parse default-ns s)
  (with-xmltm-environment
   env default-ns
   (let sub ((env env)
	     (t (parse-html s)))
     (cond ((string? t) t)
	   ((sxml-top-node? t) `(*TOP* ,@(map (cut sub env <>) (cdr t))))
	   ((sxml-control-node? t) t)
	   (else (ns-import-element sub env t))))))

(define (null->false x) (if (null? x) #f x))

(define (ns-import-element sub env t)
  (consume-xmlns
   env (sxml-attr-list t)
   (lambda (env attrs)
     ((cut sxml-set-name <> (ns-import-name env #t (sxml-name t)))
      ((cut sxml-set-attrlist <> (null->false (ns-import-attrs env attrs)))
       (sxml-set-content t (map (cut sub env <>)
				(sxml-content t))))))))

(define (ns-import-attrs env attrs)
  ((cut map <> attrs)
   (lambda (attr)
     (list (ns-import-name env #f (first attr))
	   (second attr)))))

(define (ns-import-name env use-default? name)
  (receive (ns-id ncname) (sxml-split-name name)
    (let ((ns-uri (cond (ns-id (environment-ref* env (string->symbol ns-id)))
			(use-default? (environment-ref env *default*))
			(else ""))))
      (string->symbol
       (string-append
	;; FIXME: user namespace prefix list should be extensible
	(cond ((== ns-uri xmlns-uri-xhtml) "h:")
	      ((== ns-uri xmlns-uri-mathml) "m:")
	      ((== ns-uri xmlns-uri-xml) "x:")
	      ((string-null? ns-uri) "")
	      (else (string-append ns-uri ":")))
	ncname)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; htmltm environment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-macro (with-htmltm-environment env . body)
  `(let ((,env (environment)))
     (with-environment ,env ((preserve-space? #f))
       ,@body)))

(define (htmltm-preserve-space? env)
  (environment-ref env preserve-space?))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HTML tools
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (htmltm-string s)
  (utf8->cork s))

(define (shtml-attr-non-null as att)
  ;; Get an HTML attribute or false if the attribute is absent, is not set, or
  ;; is set to the empty string.
  (and-let* ((l (assoc att as))
	     ((pair? (cdr l)))
	     ((not (string-null? (second l)))))
    (second l)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Label constructors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (attr->label a name)
  (let ((val (shtml-attr-non-null a name)))
    (and val `(label ,(htmltm-string val)))))

(define (xmltm-label-decorate a name t)
  (let ((label (attr->label a name)))
    (if label
	(stm-insert-first-data t label)
	t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Whitespace handling
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (htmltm-space-element env l)
  ;; remove string nodes in the sxml node-list @l.
  (sxml-filter-element-content l))

(define (htmltm-collapse-spaces s)
  ;; normalize (convert to #\space) and collapse whitespace in s.
  (list->string (string-fold-right htmltm-collapse-spaces/kons '() s)))

(define (htmltm-collapse-spaces/kons kar kdr)
  (cond ((not (char-whitespace? kar)) (cons kar kdr))
	((null? kdr) (list #\space))
	((char-whitespace? (first kdr)) kdr)
	(else (cons #\space kdr))))

(define (htmltm-space-mixed env l)
  ;; remove heading and trailing spaces, and collapses whitespaces in sxml node
  ;; list @l. Correctly merges consecutive string nodes in @l.
  (cond ((null? l) '())
	((htmltm-preserve-space? env) (htmltm-space-preserve l))
	(else (let ((l2 (list-fold-right htmltm-space-mixed/kons #f l)))
		(if (string? (first l2))
		    (cons (htmltm-collapse-spaces (string-trim (car l2)))
			  (cdr l2))
		    l2)))))

(define (htmltm-space-mixed/kons kar kdr)
  (cond ((not kdr)			; kar is last node
	 (if (string? kar)
	     (list (string-trim-right kar))
	     (list kar)))
	((string? kar)
	 (if (string? (first kdr))
	     (cons (string-append kar (car kdr)) (cdr kdr))
	     (cons kar kdr)))
	((string? (first kdr))
	 (cons kar (cons (htmltm-collapse-spaces (car kdr)) (cdr kdr))))
	(else
	 (cons kar kdr))))

(define (htmltm-space-preformatted env l)
  ;; Drop newline at start and end of l if present.
  ;; Assumes the parser did newline normalization.
  ;; Convert other newlines to <p> elements.
  (cond ((null? l) '())
	((htmltm-preserve-space? env) (htmltm-space-preserve l))
	(else (htmltm-space-preserve (htmltm-space-pre-first
				      (htmltm-space-pre-last l))))))

(define (htmltm-space-pre-first l)
  (let ((x (first l)))
    (if (and (string? x) (string-starts? x "\n")) ;; WARNING: \n is not R5RS
	(cons (string-drop x 1) (cdr l))
	l)))

(define (htmltm-space-pre-last l)
  (let ((x (last l)))
    (if (and (string? x) (string-ends? x "\n")) ;; WARNING: \n is not R5RS
	(rcons (but-last l) (string-drop-right x 1))
	l)))

(define (htmltm-space-preserve l)
  ;; Convert newlines to <p> elements.
  (append-map htmltm-space-preserve/sub l))

(define (htmltm-space-preserve/sub x)
  (if (string? x)
      (let ((l (string-split-lines x)))
	(if (null? (cdr l)) l (map (lambda (x) `(h:p ,x)) l)))
      (list x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Producing handlers for dispatch table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (htmltm-handler model kind method args->serial)
  ;; Produce a handler for the htmltm-methods% table
  ;;  model: content model category, for whitespace handling
  ;;         :empty -- element is defined to be empty, do not change contents.
  ;;         :element -- text nodes are ignored
  ;;           TODO: might fallback to 'mixed' if some text is present
  ;;         :mixed -- drop heading and trailing whitespaces, normalize and
  ;;           collapse internal whitespaces.
  ;;         :pre -- only drop newline at ends and switch to preserved spaces
  ;;           mode.
  ;;   kind: either :block or :inline, how is the element rendered
  ;; method: <procedure> to convert the element content to a node-list.
  ;;             @method will be passed the contents of html element.
  ;;         <string> name of a unary macro to contain the converted contents
  ;;         (<symbol> ...) use contents converted to a serial as the last
  ;;             element of this structure.
  ;;          <node list> convert to this literal, ignoring element contents.
  ;; args->serial: used iff @method matches <string> or (<symbol> ...).
  ;;             Function to convert a list of sxml nodes to a stm serial.
  ;;
  ;; The handler takes care of whitespace cleaning and (except when an
  ;; <procedure> method is provided) insert label for id attributes.
  ;;
  ;; WARNING: If a list element has an id attribute, the converted result will
  ;; not be correctly converted back by tmhtml because the first item marker
  ;; will be preceded by a label.
  ;;
  ;; *Preconditions*
  ;; List environments must have :block kind.
  ;; *Postconditions*
  ;; The handler must return a texmacs node list given a sxhtml node list.
  ;; Its result may be the empty list, a list of line-structures, or a list
  ;; containing a single document node.
  ;; *Invariants*
  ;; If @kind is :block, the handler will always return a single 'document'.
  ;; If @kind is :inline and @method is <string> or (with ...), the handler
  ;; will return a 'document' iff the serial built from the converted contents
  ;; of the html element is a 'document'.
    
  (let ((clean (cond ((eq? model :empty) (lambda (env c) c))
		     ((eq? model :element) htmltm-space-element)
		     ((eq? model :mixed) htmltm-space-mixed)
		     ((eq? model :pre) htmltm-space-preformatted)
		     (else (error "Bad model: " model))))
	(proc-alist
	 (cond ((eq? kind :inline)
		`((:procedure . ,htmltm-handler/procedure/inline)
		  (:environment . ,htmltm-handler/environment/inline)
		  (:literal . ,htmltm-handler/literal/inline)))
	       ((eq? kind :block)
		`((:procedure . ,htmltm-handler/procedure/block)
		  (:environment . ,htmltm-handler/environment/block)
		  (:literal . ,htmltm-handler/literal/block)))
	       (error "Bad kind: " kind))))

    (define (proc key) (cdr (assq key proc-alist)))
    (define (make-handler proc . extra)
      (lambda (env a c)
	(let ((cc (clean env c)))
	  (with-environment env
	      ((preserve-space? (or (eq? model :pre)
				    (htmltm-preserve-space? env))))
	    (apply proc env a cc extra)))))

    (cond ((procedure? method)
	   (make-handler (proc :procedure) method))
	  ((or (string? method) (symbol? (first method)))
	   (make-handler (proc :environment)
			 (if (string? method) ; this is ugly, should be removed
			     (if (stm-primitive? (string->symbol method))
				 `(expand ,method)
				 `(,(string->symbol method)))
			     method)
			 args->serial))
	   (else (cut (proc :literal) <> <> <> (stm-serial method))))))

(define (htmltm-handler/procedure/inline env a c proc)
  (proc env a c))

(define (htmltm-handler/procedure/block env a c proc)
  (list (stm-unary-document (htmltm-serial
			     (htmltm-preserve-space? env)
			     (proc env a c)))))

(define (htmltm-handler/literal/inline env a c literal)
  (list (xmltm-label-decorate a 'id literal)))

(define (htmltm-handler/literal/block env a c literal)
  (list (xmltm-label-decorate a 'id (stm-unary-document literal))))

(define (htmltm-handler/environment/inline env a c head args->serial)
  (list (let ((t (xmltm-label-decorate a 'id (args->serial env c))))
	  (if (stm-document? t)
	      `(document (,@head ,(stm-remove-unary-document t)))
	      `(,@head ,t)))))

(define (htmltm-handler/environment/block env a c head args->serial)
  (let ((make-block (if (stm-block-environment? head)
			stm-unary-document noop))
	(make-list (if (stm-list-environment? head)
		       htmltm-list-glue noop))
	(make-label (if (stm-section-environment? head)
			(lambda (x)
			  (xmltm-label-decorate a 'id (rcons head x)))
			(lambda (x)
			  (rcons head (xmltm-label-decorate a 'id x))))))
    (list `(document ,(make-label (make-list (make-block
					      (args->serial env c))))))))

(define (htmltm-list-glue x)
  ;; assert (stm-document? x)
  (cons 'document
	(receive (first kdr)
          (car+cdr (list-fold-right htmltm-list-glue/kons '(#f . ()) (cdr x)))
	  (htmltm-list-glue/flush first kdr))))
    
(define (htmltm-list-glue/kons kar next+kdr)
  (cons kar (receive (next kdr) (car+cdr next+kdr)
	      (if (stm-list-marker? (stm-first-data kar))
		  (cond ((not next) (list kar))
			((stm-list-marker? (stm-first-data next))
			 (cons kar kdr))
			(else
			 (cons
			  ; WARNING: no extra whitespace cleanup here
			  (htmltm-serial #t (list kar next)) kdr)))
		  ;; previous item may be a marker, do not stack kar
		  ;; but if next is not a marker, stack it now
		  (htmltm-list-glue/flush next kdr)))))

(define (htmltm-list-glue/flush line stack)
  (cond ((not line) stack)
        ((stm-list-marker? (stm-first-data line)) stack)
	(else (cons line stack))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Producing handlers for dispatch table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (mathtm-handler model method)
  ;;  model: content model category
  ;;         :empty -- element defined to be empty
  ;;         :element -- text node are ignored
  ;;  TODO: MathML error if actual content do not match model.
  ;;         :mixed -- drop heading and trailing whitespaces, normalize and
  ;;           collapse internal whitespaces.
  ;;  method: <procedure> to convert the element content to a node-list.
  (if (not (in? model '(:empty :element :mixed)))
      (error "Bad model: " model))
  (if (not (procedure? method))
      (error "Bad method: " method))
  (let ((clean (cond ((eq? model :empty) (lambda (env c) c))
		     ((eq? model :element) htmltm-space-element)
		     ((eq? model :mixed) htmltm-space-mixed))))
    (let ((proc method))
      (lambda (env a c)
	(mathtm-handler/procedure
	 env a (clean env c) proc)))))

(define (mathtm-handler/procedure env a c proc)
  (proc env a c))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Generic XML dispatcher
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (sxml-meta-drd-ref ns-id ncname)
  (cond ((== ns-id "h")
	 ;; FIXME: ugly interface from HTML to MathML
	 (or (drd-ref htmltm-methods% ncname)
	     (drd-ref mathtm-methods% ncname)))
	((== ns-id "m") (drd-ref mathtm-methods% ncname))
	(else #f)))

(define (sxml-dispatch x-string x-pass default-ns env t)
  ;; Generic xml dispatcher
  ;; @x-string (string env -> node-list) used to convert strings
  ;; @x-pass (method) pass method
  ;; @default-ns (string) is the internal prefix of current default namespace
  ;; NOTE: method == (env atts body -> node-list)
  (cond ((string? t) (x-string env t))
	((== (sxml-name t) '*TOP*) (x-pass env '() (sxml-content t)))
	((sxml-control-node? t) '())
	(else (let ((ns-id (or (sxml-name->ns-id (sxml-name t)) default-ns))
		    (ncname (string->symbol (sxml-ncname t))))
		(cond ((sxml-meta-drd-ref ns-id ncname)
		       => (lambda (m)
			    (m env (sxml-attr-list t) (sxml-content t))))
		      (else
		       (x-pass env (sxml-attr-list t) (sxml-content t))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Special serial constructors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; For performance, we maintain the invariant that block structures are always
;; wrapped in an unary 'document'.
;;
;; To maintain this invariant, it is required that htmltm-methods which
;; introduce block structures be declared of :block type.
;;
;; htmltm-handler must also take care that 'expand' and 'with' methods which
;; take a block-structure as input have their result wrapped in a 'document'.

(define (htmltm-serial p? l)
  ;; WARNING: should find a better name, since it does more than improve speed
  (if p? (stm-serial l stm-document?)
      (stm-serial l stm-document? htmltm-make-line htmltm-make-concat)))

(define (htmltm-make-line l)
  ;; Trim whitespaces at ends of lines.
  ;; Drop lines which are empty after simplification.
  (stm-concat l htmltm-make-line/concat))

(define (htmltm-make-line/concat l)
  ;; FIXME: empty lines are not always dropped...
  ;; TODO: preserve empty lines if creator is TeXmacs.
  (let ((x (stm-list->concat (stm-line-trim-both l))))
    (if (== "" x) #f x)))

(define (htmltm-make-concat l)
  ;; Trim whitespace around line-breaks.
  ;; Invisible nodes between line-breaks and whitespaces are moved.
  ;;
  ;; TODO: remove whitespaces surrounding invisibles in the middle of concat
  (stm-list->concat
   (if (null? l) '()
       (let ((line-lists (stm-parse-lines l)))
	 (if (null? (cdr line-lists)) l
	     (stm-unparse-lines
	      `(,(stm-line-trim-right (first line-lists))
		,@(map stm-line-trim-both (cDdr line-lists))
		,(stm-line-trim (last line-lists)))))))))

(define (mathtm-serial env l)
  ;; Except for the top-level math element, MathML produce only inlines.
  ;; Collapse whitespaces.
  ;; TODO: consolidate with htmltm-serial
  (stm-concat l htmltm-make-concat))
